#ifndef FREEHDL_KERNEL_CLASSES_H
#define FREEHDL_CLASSES_H

class process_base;
class acl;
class reader_info;
class sig_info_base;
class kernel_class;
class driver_info;
class wait_info;
class user_interface;


#endif
