#ifndef FREEHDL_KERNEL_H
#define FREEHDL_KERNEL_H

#include <freehdl/kernel-flags.hh>
#include <freehdl/kernel-acl.hh>
#include <freehdl/kernel-map-list.hh>
#include <freehdl/kernel-name-stack.hh>
#include <freehdl/kernel-process-base.hh>
#include <freehdl/kernel-delayed-procedure-base.hh>
#include <freehdl/kernel-winfo-item.hh>
#include <freehdl/kernel-sig-info.hh>
#include <freehdl/kernel-kernel-class.hh>
#include <freehdl/kernel-sigacl-list.hh>
#include <freehdl/kernel-wait-info.hh>
#include <freehdl/kernel-reader-info.hh>
#include <freehdl/kernel-driver-info.hh>
#include <freehdl/kernel-handle.hh>
#include <freehdl/kernel-attributes.hh>
#include <freehdl/kernel-resolver-descriptor.hh>

#endif
